function ContentType(mimeType, charset) {
  this.mimeType = mimeType
  this.charset = charset
}

function __null_aware_to_string(value) {
  return value != null ? "" + value : value
}

function Variables(global) {
  this.vars = {}

  for (var key in global) {
    this.vars[global[key].getKey()] = __null_aware_to_string(global[key].getValue())
  }

  this.set = function (varName, varValue) {
    this.vars[varName] = varValue
  }

  this.get = function (varName) {
    return this.vars[varName]
  }

  this.isEmpty = function () {
    return Object.keys(this.vars).length === 0
  }

  this.clear = function (varName) {
    delete this.vars[varName]
  }

  this.clearAll = function () {
    this.vars = {}
  }
}

function HttpClient(vars, canelable, logPrinter) {
  this.tests = {}
  this.global = new Variables(vars)

  this.test = function (testName, func) {
    let exception
    try {
      throw new Error("error")
    }
    catch (e) {
      exception = __exception__
    }
    this.tests[testName] = {
      "func": func || null,
      "creationException": exception,
    }
  }

  this.assert = function (condition, message) {
    if (!condition) {
      const err = new Error(message || "Assert failed")
      err.name = "AssertionError"
      throw err
    }
  }

  this.log = function (text) {
    logPrinter.accept(text + "\n")
  }

  this.exit = function () {
    canelable.accept(this)
  }
}

function ResponseHeaders(headers) {
  this.headers = headers

  this.valueOf = function (headerName) {
    const len = this.headers.length
    for (let i = 0; i < len; i++) {
      if (headerName == this.headers[i].name) {
        return __null_aware_to_string(this.headers[i].value)
      }
    }
  }

  this.valuesOf = function (headerName) {
    let values = []
    let len = this.headers.length
    for (let i = 0; i < len; i++) {
      if (headerName == headers[i].name) {
        values.push(__null_aware_to_string(headers[i].value))
      }
    }
    return values
  }

  this.toJSON = function () {
    const obj = {};
    const len = this.headers.length

    for (let i = 0; i < len; i++) {
      let key = "" + headers[i].name;
      let headerValue = obj[key];
      if (headerValue === undefined) {
        obj[key] = headers[i].value
      }
      else if (headerValue instanceof Array) {
        headerValue.push(headers[i].value)
      }
      else {
        obj[key] = [headerValue, headers[i].value]
      }
    }

    return obj;
  }
}

function TextStreamResponse(
  linesSubscriberConsumer,
  messagesSubscriberConsumer,
  isJson
) {
  this.onEachLine = function (subscriber, onFinish) {
    const finalSubscriber = isJson
      ? function (line, unsubscribe) {
        let newLine
        try {
          newLine = JSON.parse(line)
        }
        catch (e) {
          newLine = line
        }
        subscriber(newLine, unsubscribe)
      }
      : subscriber
    linesSubscriberConsumer.accept(finalSubscriber, onFinish)
  }

  this.onEachMessage = function (subscriber, onFinish) {
    const finalSubscriber = isJson ?
      function (message, unsubscribe, output) {
        let newMessage
        try {
          newMessage = JSON.parse(message)
        }
        catch (e) {
          newMessage = message
        }
        subscriber(newMessage, unsubscribe, output)
      }
      : subscriber
    messagesSubscriberConsumer.accept(finalSubscriber, onFinish)
  }
}

function RequestHeader(h) {
  this.__headerValue = h.getSubstitutedValue();
  this.name = __null_aware_to_string(h.name);
  this.value = function () {
    return __null_aware_to_string(this.__headerValue.getValue())
  }
}

function Request(requestElementsProvider) {
  this.__url = requestElementsProvider.getSubstitutedUrl();
  this.__body = requestElementsProvider.getSubstitutedBody();

  this.url = function() {
    return __null_aware_to_string(this.__url.getValue())
  }

  this.body = function() {
    return __null_aware_to_string(this.__body.getValue())
  }

  this.environment = {
    __env: requestElementsProvider,
    get: function (name) {
      return __null_aware_to_string(this.__env.getEnvironmentVariable(name));
    }
  }

  this.variables = {
    __provider: requestElementsProvider,
    get: function (name) {
      return __null_aware_to_string(this.__provider.getSessionVariable(name))
    }
  }

  this.headers = {
    __headers: requestElementsProvider.getHeaders(),
    all: function () {
      const iterator = this.__headers.getValue().iterator()
      const result = []
      while (iterator.hasNext()) {
        let h = iterator.next()
        result.push(new RequestHeader(h));
      }
      return result;
    },
    findByName: function(name) {
      const result = this.__headers.getValue().getHeader(name)
      return result !== null ? new RequestHeader(result) : null
    }
  }
}

const jsHandler = function (
  isJson,
  global,
  responseBody,
  statusCode,
  mimeType,
  charset,
  headers,
  cancelable,
  logPrinter,
  linesSubscriberConsumer,
  messagesSubscriberConsumer,
  isStreaming,
  requestElementsProvider
) {
  const client = new HttpClient(global, cancelable, logPrinter)

  let userVisibleBody
  try {
    if (isStreaming) {
      userVisibleBody = new TextStreamResponse(linesSubscriberConsumer, messagesSubscriberConsumer, isJson)
    }
    else {
      userVisibleBody = isJson ? JSON.parse(responseBody) : responseBody
    }
  }
  catch (e) {
    userVisibleBody = responseBody
  }

  const response = {
    body: userVisibleBody,
    status: statusCode,
    headers: new ResponseHeaders(headers),
    contentType: new ContentType(mimeType, charset)
  }

  const request = new Request(requestElementsProvider);
  try {
    //<CUSTOM_SCRIPT_PLACEHOLDER>

    return client
  }
  catch (e) {
    if (e instanceof JavaException && e.message.startsWith("com.intellij.httpClient.http.request.run.rhino.HttpClientRhinoResponseHandler$ExitScriptException")) {
      return client
    }
    else {
      if (e.rhinoException === undefined) {
        e.rhinoException = __exception__ // see /org/mozilla/javascript/ScriptRuntime.java:4118
      }
      throw e
    }
  }
}

function ResponseHandlerErrorWrapper(error) {
  this.__wrappedError__ = error
}

function AssertionError(message) {
  this.__assertion_message__ = message
}